﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace SerialSave
{
    public partial class Form1 : Form
    {
        SerialPort serialPort = null;
        Boolean saveFlag = false;
        StreamWriter sw = null;

        Boolean autoDateFilename = false;
        String saveFolder;

        bool systeminput = false;
        bool limitNum = false;
        bool limitSec = false;
        long maxNum = 0;
        long maxMillis = 0;
        long startMillis = 0;
        long nowNum = 0;
        long nowMillis = 0;

        DateTime prevDateTime;

        String stringIn = null;
        String stringOut = null;

        bool writeDate = false;
        String textAnswer = null;

        private delegate void SafeCallDelegate(string msg);

        private void appendMsg(String msg)
        {
            if (textBoxMsg.InvokeRequired)
            {
                var d = new SafeCallDelegate(appendMsg);
                textBoxMsg.Invoke(d, new object[] { msg });
            }
            else
            {
                textBoxMsg.AppendText(msg);
                textBoxMsg.AppendText("\r\n");

                if (textBoxMsg.Lines.Length > 100)
                {
                    while (textBoxMsg.Lines.Length > 90)
                    {
                        textBoxMsg.Text = textBoxMsg.Text.Remove(0, textBoxMsg.Text.IndexOf("\r\n") + 2);
                    }

                    textBoxMsg.ScrollToCaret();
                }

            }
        }

        /*
        public void showLine(string line)
        {
            textBoxLast.Invoke(new UpdateTextBoxLastHandler(UpdateTextBoxLast), line);
        }
        */

        private void showLine(String msg)
        {
            if (textBoxLast.InvokeRequired)
            {
                var d = new SafeCallDelegate(showLine);
                textBoxLast.Invoke(d, new object[] { msg });
            }
            else
            {
                textBoxLast.Text = msg;
            }
        }


        public Form1()
        {
            InitializeComponent();

            buttonRefresh_Click(null, null);

            systeminput = true;

            try
            {
                textBoxBaudRate.Text = Properties.Settings.Default.BaudRate;
                if (textBoxBaudRate.Text.Length < 1)
                {
                    textBoxBaudRate.Text = "115200";
                }

                textBoxLimitNum.Text = Properties.Settings.Default.LimitLine.ToString();
                checkBoxLimitNum.Checked = Properties.Settings.Default.LimitLineFlag;

                textBoxLimitSec.Text = Properties.Settings.Default.LimitSec.ToString();
                checkBoxLimitSec.Checked = Properties.Settings.Default.LimitSecFlag;

                textBoxTargetIn.Text = Properties.Settings.Default.IncludeKeyword.ToString();
                checkBoxTargetIn.Checked = Properties.Settings.Default.IncludeKeywordFlag;

                textBoxTargetOut.Text = Properties.Settings.Default.ExcludeKeyword.ToString();
                checkBoxTargetOut.Checked = Properties.Settings.Default.ExcludeKeywordFlag;

                checkBoxWriteDate.Checked = Properties.Settings.Default.DateFlag;

                textBoxAnswer.Text = Properties.Settings.Default.Answer.ToString();
                checkBoxAnswer.Checked = Properties.Settings.Default.AnswerFlag;

                checkBoxPrintWriteStatus.Checked = Properties.Settings.Default.ShowProc;

                checkBoxAutoDateFilename.Checked = Properties.Settings.Default.AutoStart;
                if (checkBoxAutoDateFilename.Checked)
                {
                    prevDateTime  = DateTime.Now;
                    for (int i = 0; i < listBoxComs.Items.Count; i++)
                    {
                        string tname = listBoxComs.Items[i].ToString();
                        if (tname.Equals(Properties.Settings.Default.ComPortName))
                        {
                            listBoxComs.SelectedIndex = i;
                            break;
                        }
                    }
                    comConnect();

                    textBoxFilename.Text = Properties.Settings.Default.TargetFolder;
                    autoDateFilename = true;
                    saveFolder = Properties.Settings.Default.TargetFolder;
                    setAutoFilename();
                    saveFlag = true;

                    checkBoxLimitNum.Enabled = false;
                    checkBoxLimitSec.Enabled = false;
                    checkBoxTargetIn.Enabled = false;
                    checkBoxTargetOut.Enabled = false;
                    checkBoxWriteDate.Enabled = false;
                    checkBoxAnswer.Enabled = false;
                    buttonStartStop.Text = "保存停止";
                }

            }
            catch (Exception ex)
            {
                appendMsg("Fail to recover previous settings.");
            }


            systeminput = false;


            var task = Task.Run(() =>
            {
                while (true)
                {
                    if ((serialPort != null) && (serialPort.IsOpen))
                    {
                        try
                        {
                            string s = serialPort.ReadLine();

                            if (stringIn != null)
                            {
                                if (s.IndexOf(stringIn) < 0)
                                {
                                    continue;
                                }
                            }

                            if (stringOut != null)
                            {
                                if (s.IndexOf(stringOut) >= 0)
                                {
                                    continue;
                                }
                            }

                            DateTime nowDateTime = DateTime.Now;

                            string sData = null;
                            if (writeDate)
                            {
                                if (textAnswer != null)
                                {
                                    sData = nowDateTime.ToString("yyy/MM/dd HH:mm:ss.fff") + "," + textAnswer + "," + s;
                                }
                                else
                                {
                                    sData = nowDateTime.ToString("yyy/MM/dd HH:mm:ss.fff") + "," + s;
                                }
                            }
                            else
                            {
                                if (textAnswer != null)
                                {
                                    sData = textAnswer + "," + s;
                                }
                                else
                                {
                                    sData = s;
                                }
                            }


                            if (saveFlag)
                            {
                                showLine("W : " + sData);

                                //
                                if (nowDateTime.Day > prevDateTime.Day)
                                {
                                    // 日付をまたいだ
                                    setAutoFilename();
                                }

                                if (sw != null)
                                {
                                    sw.Write(sData);

                                    nowNum++;
                                    if ((nowNum % 100) == 0)
                                    {
                                        sw.Flush();
                                    }

                                    nowMillis = currentTimeMillis();

                                    long difMillis = nowMillis - startMillis;

                                    if (checkBoxPrintWriteStatus.Checked)
                                    {
                                        appendMsg("num " + nowNum + ", " + difMillis + "ms : " + sData);
                                    }

                                    if ((limitNum && (nowNum > maxNum)) || (limitSec && ((difMillis > maxMillis))))
                                    {
                                        saveFlag = false;
                                        if (sw != null)
                                        {
                                            sw.Flush();
                                            sw.Close();
                                        }
                                        buttonStartStop.Text = "保存開始";
                                        checkBoxLimitNum.Enabled = true;
                                        checkBoxLimitSec.Enabled = true;
                                        checkBoxTargetIn.Enabled = true;
                                        checkBoxTargetOut.Enabled = true;
                                        checkBoxWriteDate.Enabled = true;
                                        checkBoxAnswer.Enabled = true;
                                    }
                                } else
                                {
                                    appendMsg("sw not correct");
                                }

                            } else
                            {
                                // textBoxLast.Invoke(new UpdateTextBoxLastHandler(UpdateTextBoxLast), sData);
                                showLine(sData);
                            }

                            prevDateTime = nowDateTime;
                        } catch (Exception e)
                        {

                        }
                    }

                    Task.Delay(10);
                }
            });
        }

        ~Form1()
        {

        }

        private void exitProc()
        {
            try
            {
                if (sw != null)
                {
                    sw.Flush();
                    sw.Close();
                }
            } catch (Exception e)
            {
                ;
            }

            Properties.Settings.Default.BaudRate = textBoxBaudRate.Text;

            Properties.Settings.Default.LimitLineFlag = checkBoxLimitNum.Checked;
            if (checkBoxLimitNum.Checked)
            {
                Properties.Settings.Default.LimitLine = int.Parse(textBoxLimitNum.Text);
            }

            Properties.Settings.Default.LimitSecFlag = checkBoxLimitSec.Checked;
            if (checkBoxLimitSec.Checked)
            {
                Properties.Settings.Default.LimitSec = int.Parse(textBoxLimitSec.Text);
            }

            Properties.Settings.Default.IncludeKeywordFlag = checkBoxTargetIn.Checked;
            if (checkBoxTargetIn.Checked)
            {
                Properties.Settings.Default.IncludeKeyword = textBoxTargetIn.Text;
            }

            Properties.Settings.Default.ExcludeKeywordFlag = checkBoxTargetOut.Checked;
            if (checkBoxTargetOut.Checked)
            {
                Properties.Settings.Default.ExcludeKeyword = textBoxTargetOut.Text;
            }

            Properties.Settings.Default.DateFlag = checkBoxWriteDate.Checked;

            Properties.Settings.Default.AnswerFlag = checkBoxAnswer.Checked;
            if (checkBoxAnswer.Checked)
            {
                Properties.Settings.Default.Answer = textBoxAnswer.Text;
            }

            Properties.Settings.Default.AutoStart = checkBoxAutoDateFilename.Checked;
            Properties.Settings.Default.TargetFolder = saveFolder;

            Properties.Settings.Default.ShowProc = checkBoxPrintWriteStatus.Checked;

            try
            {
                Properties.Settings.Default.ComPortName = listBoxComs.SelectedItem.ToString();
            } catch (Exception ex)
            {
                ;
            }

            Properties.Settings.Default.Save();
        }

        public delegate void UpdateTextBoxLastHandler(string msg);

        public void UpdateTextBoxLast(string msg)
        {
            textBoxLast.Text = msg;
        }


        private void buttonFileSelect_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfDialog = new SaveFileDialog();
            sfDialog.Filter = "CSVファイル(*.csv)|*.csv|すべてのファイル(*.*)|*.*";

            if (sfDialog.ShowDialog() == DialogResult.OK)
            {
                if (File.Exists(sfDialog.FileName))
                {
                    DialogResult result = MessageBox.Show("ファイルに追記しますか？",
                                                            "質問",
                                                            MessageBoxButtons.YesNoCancel,
                                                            MessageBoxIcon.Exclamation,
                                                            MessageBoxDefaultButton.Button2);
                    if (result == DialogResult.Yes)
                    {
                        textBoxFilename.Text = sfDialog.FileName;
                    } else
                    {
                        textBoxMsg.AppendText("キャンセルしました。\r\n");
                    }
                } else {
                    textBoxFilename.Text = sfDialog.FileName;
                }
            }
        }

        private static readonly DateTime Jan1st1970 = new DateTime
    (1970, 1, 1, 0, 0, 0, DateTimeKind.Utc);

        public static long currentTimeMillis()
        {
            return (long)(DateTime.UtcNow - Jan1st1970).TotalMilliseconds;
        }


        private void buttonStartStop_Click(object sender, EventArgs e)
        {
            if (saveFlag)
            {
                // trueなので保存停止
                saveFlag = false;
                if (sw != null)
                {
                    sw.Close();
                }
                buttonStartStop.Text = "保存開始";
                checkBoxLimitNum.Enabled = true;
                checkBoxLimitSec.Enabled = true;
                checkBoxTargetIn.Enabled = true;
                checkBoxTargetOut.Enabled = true;
                checkBoxWriteDate.Enabled = true;
                checkBoxAnswer.Enabled = true;
            }
            else
            {
                // falseなので保存開始
                checkBoxLimitNum.Enabled = false;
                checkBoxLimitSec.Enabled = false;
                checkBoxTargetIn.Enabled = false;
                checkBoxTargetOut.Enabled = false;
                checkBoxWriteDate.Enabled = false;
                checkBoxAnswer.Enabled = false;

                if ((textBoxFilename.Text != null) && (textBoxFilename.Text.Length > 1))
                {
                    sw = new StreamWriter(textBoxFilename.Text, true);
                    if (sw != null)
                    {
                        if (writeDate)
                        {
                            if (textAnswer != null)
                            {
                                sw.Write("datetime,answer,data");
                            }
                            else
                            {
                                sw.Write("datetime,data");
                            }
                        } else
                        {
                            if (textAnswer != null)
                            {
                                sw.Write("answer,data");
                            }
                            else
                            {
                                sw.Write("data");
                            }

                        }
                        saveFlag = true;
                        buttonStartStop.Text = "保存停止";
                    } else
                    {
                        textBoxMsg.AppendText("ファイルを開くことができませんでした。別のファイルを指定してください。\r\n");
                    }
                }
                else
                {
                    textBoxMsg.AppendText("ファイルを選択してから保存開始してください。\r\n");
                }
            }
        }

        private void buttonExit_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void buttonRefresh_Click(object sender, EventArgs e)
        {
            listBoxComs.Items.Clear();

            string[] ports = SerialPort.GetPortNames();
            foreach (string port in ports)
            {
                listBoxComs.Items.Add(port);
            }

            if (listBoxComs.Items.Count > 0)
            {
                listBoxComs.SetSelected(0, true);
            } else
            {
                textBoxMsg.AppendText("シリアルポートがありません。USBケーブルを接続してください。\r\n");
            }
        }

        private void comConnect()
        {
            if (listBoxComs.SelectedIndex < 0)
            {
                textBoxMsg.AppendText("シリアルポートを選択してください。\r\n");
                return;
            }

            if (serialPort != null)
            {
                if (serialPort.IsOpen)
                {
                    serialPort.Close();
                }
            }

            serialPort = new SerialPort();
            serialPort.PortName = listBoxComs.SelectedItem.ToString();
            // serialPort.BaudRate = 1552000;
            serialPort.BaudRate = int.Parse(textBoxBaudRate.Text);
            serialPort.DataBits = 8;
            serialPort.Parity = Parity.None;
            serialPort.StopBits = StopBits.One;
            serialPort.Handshake = Handshake.None;
            // serialPort.Encoding = Encoding.UTF8;
            // serialPort.RtsEnable = true;

            // serialPort.DataReceived += new SerialDataReceivedEventHandler(SerialPort_DataReceived);

            try
            {
                serialPort.Open();
                textBoxMsg.AppendText("" + serialPort.PortName + " connect");
            }
            catch (Exception ex)
            {
                textBoxMsg.AppendText("接続できませんでした。USBケーブルを再接続してみてください。\r\n");
            }

        }

        private void buttonConnect_Click(object sender, EventArgs e)
        {
            comConnect();
        }

        private void checkBoxLimitNum_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxLimitNum.Checked)
            {
                limitNum = true;
                maxNum = long.Parse(textBoxLimitNum.Text);
                nowNum = 0;
            } else
            {
                limitNum = false;
            }
        }

        private void checkBoxLimitSec_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxLimitSec.Checked)
            {
                limitSec = true;
                maxMillis = long.Parse(textBoxLimitSec.Text) * 1000L;
                nowMillis = 0;
                startMillis = currentTimeMillis();
            } else
            {
                limitSec = false;
            }
        }

        private void checkBoxTargetIn_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxTargetIn.Checked)
            {
                stringIn = textBoxTargetIn.Text;
            } else
            {
                stringIn = null;
            }
        }

        private void checkBoxTargetOut_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxTargetOut.Checked)
            {
                stringOut = textBoxTargetOut.Text;
            } else
            {
                stringOut = null;
            }
        }

        private void checkBoxWriteDate_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxWriteDate.Checked)
            {
                writeDate = true;
            } else
            {
                writeDate = false;
            }
        }

        private void checkBoxAnswer_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxAnswer.Checked)
            {
                textAnswer = textBoxAnswer.Text;
            } else
            {
                textAnswer = null;
            }
        }

        private void buttonSerialClose_Click(object sender, EventArgs e)
        {
            if (serialPort != null)
            {
                if (serialPort.IsOpen)
                {
                    serialPort.Close();
                }
            }
        }

        private void Form1_Load(object sender, EventArgs e)
        {
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if ((serialPort != null) && (serialPort.IsOpen))
            {
                serialPort.Close();
            }

            exitProc();

        }

        private void setAutoFilename()
        {
            if (sw != null)
            {
                sw.Flush();
                sw.Close();
            }

            String filename = saveFolder + "\\" + DateTime.Now.ToString("yyyMMdd") + ".csv";
            if (!File.Exists(filename))
            {
                textBoxFilename.Text = filename;
                sw = new StreamWriter(filename, true);
            } else
            {
                int i = 0;
                while (true)
                {
                    filename = saveFolder + "\\" + DateTime.Now.ToString("yyyMMdd") + "-" + i + ".csv";
                    if (!File.Exists(filename))
                    {
                        textBoxFilename.Text = filename;
                        sw = new StreamWriter(filename, true);
                        break;
                    }
                    i++;
                }
            }
            appendMsg("target save folder : " + saveFolder);
            appendMsg("target file : " + filename);
        }

        private void checkBoxAutoDateFilename_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxAutoDateFilename.Checked && (systeminput == false))
            {
                // 選択された
                FolderBrowserDialog fbd = new FolderBrowserDialog();
                fbd.Description = "フォルダを指定してください。";
                // fbd.RootFolder = Environment.SpecialFolder.Desktop;
                // fbd.SelectedPath = @"C:\Windows";
                fbd.ShowNewFolderButton = true;

                if (fbd.ShowDialog(this) == DialogResult.OK)
                {
                    saveFolder = fbd.SelectedPath;
                    setAutoFilename();
                }

                checkBoxLimitNum.Enabled = false;
                checkBoxLimitSec.Enabled = false;
                checkBoxTargetIn.Enabled = false;
                checkBoxTargetOut.Enabled = false;
                checkBoxWriteDate.Enabled = false;
                checkBoxAnswer.Enabled = false;
                buttonStartStop.Text = "保存停止";

                autoDateFilename = true;
                saveFlag = true;
            }
            else
            {
                autoDateFilename = false;
                saveFlag = false;
            }
        }
    }
}
